local Addon = _G[...]


local StatusBar = {}
do
	function StatusBar:New(parent)
		local bar = Addon.Merge(StatusBar, CreateFrame("Frame" , nil, parent))
		bar.scrollFrame = CreateFrame("ScrollFrame", nil, bar)
		bar.helper = CreateFrame("Frame", nil, bar.scrollFrame)

		bar.texture = bar.helper:CreateTexture(nil, "BACKGROUND")
		bar.texture:SetAllPoints(bar)
		bar.scrollFrame:SetScrollChild(bar.helper)
		bar:SetFillStyle("STANDARD")
		
		bar.Text = bar:CreateFontString(nil, 'ARTWORK', 'TextStatusBarText')
		bar.Text:SetAllPoints(bar)
		bar.text = bar.Text
		
		bar.color = {r = 1, g = 1, b = 1, a = 1}
		return bar
	end

	function StatusBar:GetFillStyle()
		return self.fillstyle
	end

	function StatusBar:GetMinMaxValues()
		return self.min or 0, self.max or 1
	end

	function StatusBar:GetOrientation() --"HORIZONTAL" or "VERTICAL
		return self.orientation or "HORIZONTAL"
	end

	function StatusBar:GetReverseFill()
		return self.reversefill
	end

	function StatusBar:GetRotatesTexture()
		return self.rotatestexture, self.rotatesdegrees or 90
	end

	function StatusBar:GetStatusBarColor()
		local color = self.color
		return color.r, color.g, color.b, color.a
	end

	function StatusBar:GetStatusBarTexture()
		return self.texture
	end

	function StatusBar:GetValue()
		return self.value
	end

	function StatusBar:SetFillStyle(style)
		self.fillstyle = style
		self.scrollFrame:ClearAllPoints()
		local orientation = self:GetOrientation()
		if style ~= "REVERSE" then
			self.reversefill = nil
		end
		
		if orientation == "BOTH" then
			self.scrollFrame:SetPoint("Center")
		elseif orientation == "STANDARD" then
			if style == "REVERSE" then
				self.scrollFrame:SetPoint("BottomLeft")
				self.scrollFrame:SetPoint("BottomRight")
			elseif style == "STANDARD_NO_RANGE_FILL" then
				self.scrollFrame:SetPoint("TopRight")
				self.scrollFrame:SetPoint("TopLeft")
			elseif style == "CENTER" then
				self.scrollFrame:SetPoint("Left")
				self.scrollFrame:SetPoint("Right")
			elseif style == "REVERSE" then
				self.scrollFrame:SetPoint("TopRight")
				self.scrollFrame:SetPoint("TopLeft")
			end
		else
			if style == "STANDARD" then
				self.scrollFrame:SetPoint("TopLeft")
				self.scrollFrame:SetPoint("BottomLeft")
			elseif style == "STANDARD_NO_RANGE_FILL" then
				self.scrollFrame:SetPoint("TopRight")
				self.scrollFrame:SetPoint("BottomRight")
			elseif style == "CENTER" then
				self.scrollFrame:SetPoint("Top")
				self.scrollFrame:SetPoint("Bottom")
			elseif style == "REVERSE" then
				self.scrollFrame:SetPoint("TopRight")
				self.scrollFrame:SetPoint("BottomRight")
			end
		end
		
		
		

	end

	function StatusBar:SetMinMaxValues(min, max)
		self.min = min
		self.max = max
	end

	function StatusBar:SetOrientation(orientation)
		self.orientation = orientation
		self:SetFillStyle(self:GetFillStyle())
	end

	function StatusBar:SetReverseFill(state)
		if state == false then
			state = nil
		end
		local orient = "STANDARD"
		if (state == true) or (state == nil) then
			self.reversefill = state
		end
		if state == true then
			orient = "REVERSE"
		else
			orient = "STANDARD"
		end
		StatusBar:SetFillStyle(orient)
	end

	function StatusBar:SetRotatesTexture(state, degrees)
		if state ~= true then
			state = nil
			self.texture:SetTexCoord(0, 1, 0, 1) --trick to reset the rotation
		end
		if state then
			self.rotatestexture = state
			self.texture:SetRotation(((degrees or 90) * math.pi) / 180)
		end
	end

	function StatusBar:SetStatusBarColor(r, g, b, a)
		local color = self.color
		color.r, color.g, color.b, color.a = r, g, b, a
		self.texture:SetVertexColor(r, g, b, a)
	end

	function StatusBar:SetStatusBarTexture(file, layer)
		self.statusbartexture = file
		
		self.texture:SetTexture(file)
		
		if layer then
			self.texture:SetDrawLayer(layer)
		end
	end

	local function Reset(self)
		self:SetAlpha(1)
		self.texture:SetAlpha(1)
		self:SetMinMaxValues(0, 1)
		self:SetValue(0)
		self.texture:Show()
		self:SetStatusBarColor(1, 1, 1, 1)
	end

	function StatusBar:Reset(fadeAway)
		if fadeAway then
			self.fadeInfo = {finishedFunc = function() Reset(self)  end}
			UIFrameFadeOut(self, .5, 1, 0)
			Reset(self)
		else
			Reset(self)
		end
	end

	function StatusBar:SetValue(value)
		self.value = value
		local min, max = self:GetMinMaxValues()

		if min ~= 0 then
			if min > 0 then
				max = max - min
			elseif max < 0 then
				max = max + math.abs(min)
			end
			min = 0
		end
		
		local orientation = self:GetOrientation()
		local width, height = self:GetSize()
		local fillstyle = self:GetFillStyle()
		
		local e = value/max

		if orientation == "BOTH" then
			if (fillstyle == "STANDARD_NO_RANGE_FILL") and (min == max) then
				self.scrollFrame:SetWidth(width)
				self.scrollFrame:SetHeight(height)
			else
				self.scrollFrame:SetWidth((e * width))
				self.scrollFrame:SetHeight((e * height))
			end
		elseif orientation == "HORIZONTAL" then
			if (fillstyle == "STANDARD_NO_RANGE_FILL") and (min == max) then
				self.scrollFrame:SetWidth(width)
			else
				self.scrollFrame:SetWidth((e * width))
			end
		else
			if (fillstyle == "STANDARD_NO_RANGE_FILL") and (min == max) then
				self.scrollFrame:SetHeight(height)
			else
				self.scrollFrame:SetHeight((e * height))
			end
		end
		if value >= max then
			--self:Hide()
		end
	end
end

local castingBar = {}
do
function castingBar:LoadColors()
	self.startCastColor = CreateColor(1.0, 0.7, 0.0)
	self.startChannelColor = CreateColor(0.0, 1.0, 0.0);
	self.finishedCastColor = CreateColor(0.0, 1.0, 0.0);
	self.nonInterruptibleColor = CreateColor(0.7, 0.7, 0.7);
	self.failedCastColor = CreateColor(1.0, 0.0, 0.0);

	self.finishedColorSameAsStart = true
	self.flashColorSameAsStart = true
end


function castingBar:GetEffectiveStartColor(isChannel, notInterruptible)
	if self.nonInterruptibleColor and notInterruptible then
		return self.nonInterruptibleColor;
	end	
	return isChannel and self.startChannelColor or self.startCastColor;
end

-- Fades additional widgets along with the cast bar, in case these widgets are not parented or use ignoreParentAlpha
function castingBar:AddWidgetForFade(widget)
	self.additionalFadeWidgets = self.additionalFadeWidgets or {};
	self.additionalFadeWidgets[widget] = true;
end


function castingBar:ApplyAlpha(alpha)
	self:SetAlpha(alpha);
	if self.additionalFadeWidgets then
		for widget in pairs(self.additionalFadeWidgets) do
			widget:SetAlpha(alpha);
		end
	end

end

function castingBar:ApplyShown(hide)
	if hide == true then
		self:Hide();
		if self.additionalFadeWidgets then
			for widget in pairs(self.additionalFadeWidgets) do
				widget:Hide();
			end
		end
			--self:Reset(true)
	else
		self:Show();
		if self.additionalFadeWidgets then
			for widget in pairs(self.additionalFadeWidgets) do
				widget:Show();
			end
		end
	end
end

function castingBar:GetCastingInfo(unit)
	local castType, func
	if UnitCastingInfo(unit) then
		castType, func = 'cast', UnitCastingInfo
	elseif UnitChannelInfo(unit) then
		castType, func = 'channel', UnitChannelInfo
	end
	if castType and func then
		return castType, func(unit)
	end
end

function castingBar:FinishSpell()
	if not self.finishedColorSameAsStart then
		self:SetStatusBarColor(self.finishedCastColor:GetRGB());
	end
	if ( self.Spark ) then
		self.Spark:Hide();
	end
--	if ( self.Flash ) then
--		self.Flash:SetAlpha(0.0);
--		self.Flash:Show();
--	end
	self.flash = true;
	self.fadeOut = true;
	self.casting = nil;
	self.channeling = nil;
end

function castingBar:TriggerEvent(event, ...)
	local arg1 = ...;
	local unit = self.id;
	if ( event == "PLAYER_ENTERING_WORLD" ) then
		local nameChannel = UnitChannelInfo(unit);
		local nameSpell = UnitCastingInfo(unit);
		if ( nameChannel ) then
			event = "UNIT_SPELLCAST_CHANNEL_START";
			arg1 = unit;
		elseif ( nameSpell ) then
			event = "UNIT_SPELLCAST_START";
			arg1 = unit;
		else
		    self:inishSpell();
		end
	end
	
	if ( event == "UNIT_SPELLCAST_START" ) then
		local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill, castID, notInterruptible = UnitCastingInfo(unit);
		if ( not name or (not self.showTradeSkills and isTradeSkill)) then
			self:ApplyShown(true)
			return;
		end

		local startColor = self:GetEffectiveStartColor(self, false, notInterruptible);
		self:SetStatusBarColor(startColor:GetRGB());
--		if self.flashColorSameAsStart then
		--	self.Flash:SetVertexColor(startColor:GetRGB());
--		else
		--	self.Flash:SetVertexColor(1, 1, 1);
--		end
		
		if ( self.Spark ) then
			self.Spark:Show();
		end
		
		self.value = (GetTime() - (startTime / 1000));
		self.maxValue = (endTime - startTime) / 1000;
		self:SetMinMaxValues(0, self.maxValue);
		self:SetValue(self.value);
		if ( self.text ) then
			self.text:SetText(text);
		end

		self:ApplyAlpha(1.0);
		self.holdTime = 0;
		self.casting = true;
		self.castID = castID;
		self.channeling = nil;
		self.fadeOut = nil;


		if ( self.showCastbar ) then
			self:ApplyShown()
		end

	elseif ( event == "UNIT_SPELLCAST_STOP" or event == "UNIT_SPELLCAST_CHANNEL_STOP") then
		if ( not self:IsVisible() ) then
			self:ApplyShown(true)
		end
		if ( (self.casting and event == "UNIT_SPELLCAST_STOP" and select(4, ...) == self.castID) or
		     (self.channeling and event == "UNIT_SPELLCAST_CHANNEL_STOP") ) then
			if ( self.Spark ) then
				self.Spark:Hide();
			end
--			if ( self.Flash ) then
--				self.Flash:SetAlpha(0.0);
--				self.Flash:Show();
--			end
			self:SetValue(self.maxValue);
			if ( event == "UNIT_SPELLCAST_STOP" ) then
				self.casting = nil;
				if not self.finishedColorSameAsStart then
					self:SetStatusBarColor(self.finishedCastColor:GetRGB());
				end
			else
				self.channeling = nil;
			end
			self.flash = true;
			self.fadeOut = true;
			self.holdTime = 0;
		end
	elseif ( event == "UNIT_SPELLCAST_FAILED" or event == "UNIT_SPELLCAST_INTERRUPTED" ) then
		if ( self:IsShown() and
		     (self.casting and select(4, ...) == self.castID) and not self.fadeOut ) then
			self:SetValue(self.maxValue);
			self:SetStatusBarColor(self.failedCastColor:GetRGB());
			if ( self.Spark ) then
				self.Spark:Hide();
			end
			if ( self.text ) then
				if ( event == "UNIT_SPELLCAST_FAILED" ) then
					self.text:SetText(FAILED);
				else
					self.text:SetText(INTERRUPTED);
				end
			end
			self.casting = nil;
			self.channeling = nil;
			self.fadeOut = true;
			self.holdTime = GetTime() + CASTING_BAR_HOLD_TIME;
		end
	elseif ( event == "UNIT_SPELLCAST_DELAYED" ) then
		if ( self:IsShown() ) then
			local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill, castID, notInterruptible = UnitCastingInfo(unit);
			if ( not name or (not self.showTradeSkills and isTradeSkill)) then
				-- if there is no name, there is no bar
				self:ApplyShown(true)
				return;
			end
			self.value = (GetTime() - (startTime / 1000));
			self.maxValue = (endTime - startTime) / 1000;
			self:SetMinMaxValues(0, self.maxValue);
			if ( not self.casting ) then
				self:SetStatusBarColor(self:GetEffectiveStartColor(self, false, notInterruptible):GetRGB());
				if ( self.Spark ) then
					self.Spark:Show();
				end
--				if ( self.Flash ) then
--					self.Flash:SetAlpha(0.0);
--					self.Flash:Hide();
--				end
				self.casting = true;
				self.channeling = nil;
				self.flash = nil;
				self.fadeOut = nil;
			end
		end
	elseif ( event == "UNIT_SPELLCAST_CHANNEL_START" ) then
		local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill, notInterruptible = UnitChannelInfo(unit);
		if ( not name or (not self.showTradeSkills and isTradeSkill)) then
			-- if there is no name, there is no bar
			self:ApplyShown(true)
			return;
		end

		local startColor = self:GetEffectiveStartColor(self, true, notInterruptible);
--		if self.flashColorSameAsStart then
--			self.Flash:SetVertexColor(startColor:GetRGB());
--		else
--			self.Flash:SetVertexColor(1, 1, 1);
--		end
		self:SetStatusBarColor(startColor:GetRGB());
		self.value = (endTime / 1000) - GetTime();
		self.maxValue = (endTime - startTime) / 1000;
		self:SetMinMaxValues(0, self.maxValue);
		self:SetValue(self.value);
		if ( self.text ) then
			self.text:SetText(text);
		end

		if ( self.Spark ) then
			self.Spark:Hide();
		end
		self:ApplyAlpha(1.0);
		self.holdTime = 0;
		self.casting = nil;
		self.channeling = true;
		self.fadeOut = nil;

		if ( self.showCastbar ) then
			self:ApplyShown()
			
		end
	elseif ( event == "UNIT_SPELLCAST_CHANNEL_UPDATE" ) then
		if ( self:IsShown() ) then
			local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill = UnitChannelInfo(unit);
			if ( not name or (not self.showTradeSkills and isTradeSkill)) then
				-- if there is no name, there is no bar
				self:ApplyShown(true)
				return;
			end
			self.value = ((endTime / 1000) - GetTime());
			self.maxValue = (endTime - startTime) / 1000;
			self:SetMinMaxValues(0, self.maxValue);
			self:SetValue(self.value);
		end
	elseif ( event == "UNIT_SPELLCAST_INTERRUPTIBLE" or event == "UNIT_SPELLCAST_NOT_INTERRUPTIBLE" ) then
		self:UpdateInterruptibleState(self, event == "UNIT_SPELLCAST_NOT_INTERRUPTIBLE");
	end
end

function castingBar:UpdateInterruptibleState(notInterruptible)
	if ( self.casting or self.channeling ) then
		local startColor = self:GetEffectiveStartColor(self, self.channeling, notInterruptible);
		self:SetStatusBarColor(startColor:GetRGB());
--		if self.flashColorSameAsStart then
--			self.Flash:SetVertexColor(startColor:GetRGB());
--		end
	end
end

function castingBar:UpdateCurrentCast(elapsed)
	local currentGUID = UnitGUID(self.id)
	local CurrentTime = GetTime()
	
	
	if not currentGUID then --no target, hide the bar
		return self:Reset()
	elseif self.previousGUID ~= currentGUID then --target has changed, force an update of cast info
		self.castType = nil
		self.spellID = nil
		self.startTime = nil
		self.endTime = nil
		self.spellname = nil
		self.seconds = ''
		self.stop = true
	end
	
	local castType, name, nameSubtext, text, texture, startTime, endTime, isTradeSkill, castID, notInterruptible = self:GetCastingInfo(self.id)
	if (castType == 'cast') and startTime then
		self.value = CurrentTime - (startTime / 1000)
	elseif (castType == 'channel') and endTime then
		self.value = (endTime / 1000) - CurrentTime
	else
		self.value = 0
	end
	
	
	local isCasting
	if (castType == "cast") or (castType == "channel") then --a spell is being cast
		self.showCastbar = true 
		if self.spellname ~= name then --a new spell has just started
			if castType == 'cast' then
				self:TriggerEvent('UNIT_SPELLCAST_START', name, nil, nil, castID)
			elseif castType == 'channel' then
				self:TriggerEvent('UNIT_SPELLCAST_CHANNEL_START', name, nil, nil, castID) --need to update for channeling
			end
		elseif self.endTime ~= endTime then --if the endTime changes, then the cast was delayed
			self:TriggerEvent('UNIT_SPELLCAST_DELAYED',  name, nil, nil, castID, spellID)
		end
		self.castType = castType
		self.spellID = select( 7, GetSpellInfo(name))
		self.startTime = startTime
		self.endTime = endTime
		self.spellname = name
		self.castID = castID or 0
		self.stop = nil
		isCasting = true
	end
	
	if (not isCasting) and self.endTime and self.startTime then
		self.showCastbar = nil 
		if ((CurrentTime) < ((self.endTime/1000) - .5)) then --the cast was interrupted
			self:TriggerEvent('UNIT_SPELLCAST_INTERRUPTED', self.spellname, nil, nil,  self.castID)
		else
			if self.castType == 'cast' then
				self:TriggerEvent('UNIT_SPELLCAST_STOP', self.spellname, nil, nil,  self.castID)
			elseif self.castType == 'channel' then
				self:TriggerEvent('UNIT_SPELLCAST_CHANNEL_STOP', self.spellname, nil, nil,  self.castID)
			end
		end
		self.castType = nil
		self.spellID = nil
		self.startTime = nil
		self.endTime = true
		self.spellname = nil
		self.seconds = ''
		self.stop = true
	else
		if self:IsShown()	then
			self:ApplyShown(true)
		end
	end
	
	if self.endTime and type(self.endTime) == "number" then
		self.seconds = string.format('%.1f' ,(self.endTime / 1000) - CurrentTime)
		self.time:SetText(self.seconds)
	else
		self.time:SetText("")
	end
	if self.spellname then
		self.text:SetText(name)
	end
	
    if self.value and self.value > 0 then
		self:ApplyShown()

		local min, max = self:GetMinMaxValues()
		if min ~= 0 then
			if min > 0 then
				max = max - min
			elseif max < 0 then
				max = max + math.abs(min)
			end
			min = 0
		end
		local orientation = self:GetOrientation()
		local fillstyle = self:GetFillStyle()
		if (fillstyle == "STANDARD_NO_RANGE_FILL") and (min == max) then
			self.spark:Hide()
			self.subspark:Hide()
		elseif fillstyle == "REVERSE" then
			self.spark:Hide()
			self.subspark:Show()
		elseif fillstyle == "CENTER" then
			self.spark:Show()
			self.subspark:Show()
		else
			self.spark:Show()
			self.subspark:Hide()
		end
	end
	
	if not self.value then
	--	return
	end
	
	if ( castType == "cast" ) then
		self.value = self.value + elapsed;
		if ( self.value >= self.maxValue ) then
			self:SetValue(self.maxValue);
			self:FinishSpell(self.Spark, self.Flash);
			return;
		end
		self:SetValue(self.value);

	elseif ( castType == "channel" ) then
		self.value = self.value - elapsed;
		if ( self.value <= 0 ) then
			self:FinishSpell(self.Spark, self.Flash);
			return;
		end
		self:SetValue(self.value);
	elseif ( self.fadeOut ) then
		local alpha = self:GetAlpha() - CASTING_BAR_ALPHA_STEP;
		if ( alpha > 0 ) then
			self:ApplyAlpha(alpha);
		else
			self.fadeOut = nil;
			self:ApplyShown(true)
		end
	end

	self.lastGUID = curGUID
end


end


local modName = ...
local title = "Cast"
--Power Bar
local widget = Addon:NewWidget(title, 'StatusBar')

widget.defaults = {
	basic = {
		advanced = {
			enable = true,
			tooltip = true,
		},
		position = {
			y = -13,
			x = 7,
			frameLevel = 5,
			anchor = "TopLeft",
			frameStrata = 2,
		},
		size = {
			enable = true,
			scale = 100,
			height = 16,
			width = 116,
		},
	},

	text = {
		text = {
			justifyH = 1,
			enable = true,
			file = "Friz Quadrata TT",
			justifyV = 2,
			color = {
				a = 1,
				b = 0,
				g = 1,
				r = 1,
			},
		},
		size = {
			height = 100,
			size = 9,
			width = 90,
		},
		position = {
			anchor = "Left",
			x = 0,
			y = 0,
		},
	},
	visibility = {
		spark = {
			y = 0,
			offset = 0,
		},
		background = {
			enable = true,
			file = "Raid",
			solid = true,
			color = {
				a = 0,
				b = 0,
				g = 0,
				r = 0,
			},
		},
		border = {
			hpadding = 0,
			vthickness = 30,
			vpadding = 0,
			color = {
				a = 0.5,
				b = 0,
				g = 0,
				r = 0,
			},
			file = "Border",
			hthickness = 50,
		},
		texture = {
			orientation = "HORIZONTAL",
			fillstyle = "STANDARD",
			file = "Raid",
			opacity = 100,
		},
	},
	time = {
		text = {
			enable = true,
			file = "Friz Quadrata TT",
			color = {
				a = 1,
				b = 0,
				g = 1,
				r = 1,
			},
			size = 9,
		},
		position = {
			anchor = "Right",
			x = 0,
			y = 0,
		},
	},
}
	
function widget:New(parent)
	local opac = CreateFrame('Frame', nil, parent.box)--opacity control
	local bar = self:Bind(StatusBar:New(opac)) --self:Bind(CreateFrame('StatusBar', name, opac, 'DU_CastingBarFrameTemplate'))

	Addon.Merge(castingBar, bar)

	
	
	bar:SetAllPoints(opac)
	bar.owner = parent
	bar.opac = opac
	--bar:Hide()
	--bar:SetAttribute('unit', parent.id)
	bar.Text:SetHeight(12)
	bar.text:SetParent(opac)
	bar.id = parent.id
	bar.time = 	bar.time or 	opac:CreateFontString(nil, 'OVERLAY', 'TextStatusBarText')
	bar.time:SetTextColor(1.0,1.0,1.0)
	bar.time:SetAllPoints(bar)
	
	bar.drop = 	bar.drop or CreateFrame('StatusBar', nil, bar)
	bar.drop:SetMinMaxValues(0,1)
	bar.drop:SetValue(1)
	bar.drop:SetAllPoints(true)
	bar:SetFrameLevel(3)
	bar.drop:SetFrameLevel(bar:GetFrameLevel()-2)
	
	bar.testTexture = bar.testTexture or CreateFrame('StatusBar', nil)
	bar.testTexture:SetMinMaxValues(0,1)
	bar.testTexture:SetValue(1)
	bar.testTexture:SetAllPoints(bar)
	bar.testTexture:SetFrameLevel(bar:GetFrameLevel()-2)
	
	
	bar.spark = opac:CreateTexture(nil, 'ARTWORK')
	bar.spark:SetPoint("Center")
	bar.spark :SetSize(30,30)
	bar.spark:SetTexture("Interface/CastingBar/UI-CastingBar-Spark")
	bar.spark:SetBlendMode("ADD")

	bar.subspark = opac:CreateTexture(nil, 'ARTWORK')
	bar.subspark:SetPoint("Center")
	bar.subspark :SetSize(30,30)
	bar.subspark:SetTexture("Interface/CastingBar/UI-CastingBar-Spark")
	bar.subspark:SetBlendMode("ADD")
	
	local texture = bar:CreateTexture(nil, "ARTWORK")
	texture:SetParent(Addon.HIDER)
	
	bar.Border = texture
	bar.BorderShield = texture
	bar.Icon = texture
	bar.Flash = texture
	bar.Spark = texture
--[[	
	bar.Border:SetParent(Addon.HIDER)
	bar.BorderShield:SetParent(Addon.HIDER)
	bar.Icon:SetParent(Addon.HIDER)
	bar.Flash:SetParent(Addon.HIDER)
	bar.Spark:SetParent(Addon.HIDER)
--]]	
	bar.border = {}

	bar.borderHandler = CreateFrame("Frame", nil, bar.opac)
	
	bar.border.TL 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	bar.border.L 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	bar.border.BL 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	bar.border.B 	= bar.opac:CreateTexture(nil, 'ARTWORK')	
	bar.border.BR 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	bar.border.R 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	bar.border.TR 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	bar.border.T 	= bar.opac:CreateTexture(nil, 'ARTWORK')	
	bar.border.C 	= bar.opac:CreateTexture(nil, 'ARTWORK')
	
	bar.owner = parent
	bar.title = title

	
	bar:AddWidgetForFade(bar.borderHandler)
	bar:AddWidgetForFade(bar.drop)
	bar:AddWidgetForFade(bar.text)
	bar:AddWidgetForFade(bar.time)
	bar:AddWidgetForFade(bar.texture)
	bar:AddWidgetForFade(bar.opac)
	
	for widget in pairs(bar.border) do
--		bar:AddWidgetForFade(widget)
	end

	
	
	bar:ApplyAlpha(0)
	bar.opac:Hide()
	
	
--	bar.scrollFrame:SetFrameStrata("BACKGROUND")
	
	bar.handler = parent.id
	return bar
end

function widget:Load()
	self:LoadColors()
	self:SetScript('OnShow', self.OnShow)
	self:SetScript('OnEvent', nil)
	self:Show()
end

function widget:Layout()
	if self.sets.basic.advanced.enable~= true then
		self:Hide()
		self.noUpdate = true
		return
	else
		self:Show()
		self.noUpdate = nil
	end
	self:Show()
	self:Resize()
	self:Reposition()	
	self:LayoutText()
	self:LayoutTime()
	self:SetVisibility()

	if self.SetBorder then
		self:SetBorder()
	end
	
	self:OnUpdate(1)
end


local function LookForSets(frame)
	if not frame:GetParent() then
		return nil, "Dead End"
	elseif frame:GetParent().sets then
		return frame:GetParent().sets
	else
		return LookForSets(frame:GetParent())
	end
end

function widget:Resize()
	local size = self.sets.basic.size
	
	local width, height = (size.width), (size.height)
	self.opac:SetHeight(height)
	
	local scale = size.scale/100
	self.opac:SetScale(scale)
	
	local set = LookForSets(self.opac)
	if set then
		local w = set.width
		if set.magicWidth then
			local d = w - 191
			self.opac:SetWidth((size.width  + d)/ scale)
		else
			self.opac:SetWidth(width)
		end
	else
		self.opac:SetWidth(width)
	end
end


function widget:Reposition()
	local position = self.sets.basic.position
	local scale = self.sets.basic.size.scale/100
	self.opac:ClearAllPoints()
	self.opac:SetPoint(position.anchor, self.opac:GetParent(), position.x / scale, position.y / scale)
	
	
	
	local lay = Addon.layers[position.frameStrata]
	self:SetFrameStrata(lay)
	local level = position.frameLevel + 2
	self:SetFrameLevel(level)
	

	self.drop:SetFrameLevel(position.frameLevel)
	
	local lay = Addon.layers[position.frameStrata+1]
	self.opac:SetFrameStrata(lay)
	self.opac:SetFrameLevel(position.frameLevel+1)
end


local _hori = {
	'LEFT',
	'CENTER',
	'RIGHT',
}

local _vert = {
	'TOP',
	'MIDDLE',
	'BOTTOM',
}

function widget:SetSpark()
    if self.sets.visibility.texture.rotateTexture == true then
        self.spark:SetRotation(math.rad(-90))
    else
        self.spark:SetRotation(math.rad(0))
    end
end


function widget:LayoutText()
	local text = self.text
		
	local font = self.sets.text.text
	if font.enable then
		text:Show()
	else
		text:Hide()
		return
	end

	text:SetJustifyH(_hori[font.justifyH] or font.justifyH or "CENTER")
	text:SetJustifyV(_vert[font.justifyV] or font.justifyV or "MIDDDLE")
	text:SetFont(self:GetMediaPath("font", font.file), self.sets.text.size.size or 12)
	text:SetTextColor(font.color.r, font.color.g, font.color.b, font.color.a)
	
	local w, h = self.sets.basic.size.width, self.sets.basic.size.height
	w = w * (self.sets.text.size.width/ 100)
	h = h * (self.sets.text.size.height/ 100)
	text:SetSize(w, h)
	local position = self.sets.text.position
	text:ClearAllPoints()
	text:SetPoint(position.anchor, self, position.anchor, position.x, position.y)
end

function widget:LayoutTime()
	local text = self.time
		
	local font = self.sets.time.text
	if font.enable then
		text:Show()
	else
		text:Hide()
		return
	end

	text:SetFont(self:GetMediaPath("font", font.file), font.size or 12)
	text:SetTextColor(font.color.r, font.color.g, font.color.b, font.color.a)
	
	local position = self.sets.time.position
	text:ClearAllPoints()
	text:SetPoint(position.anchor, self, position.anchor, position.x, position.y)
end


function widget:SetVisibility()
	local visibility = self.sets.visibility
	
	local background = visibility.background
    if background.enable then
        self.drop:SetStatusBarColor(background.color.r, background.color.g, background.color.b, 1-background.color.a)
    else
        self.drop:SetStatusBarTexture("")
    end

	local texture = visibility.texture
	self.isReverse = texture.fillstyle == "REVERSE"
	self:SetFillStyle(texture.fillstyle);

	self:SetOrientation(texture.orientation) -- "HORIZONTAL" or "VERTICAL"
	
		local a,b = "Top", "Bottom"
		if texture.orientation == "HORIZONTAL" then
			a,b = "Left", "Right"
		end

		self.subspark:SetPoint("Center", self.scrollFrame, a)
		self.spark:SetPoint("Center", self.scrollFrame,    b)
	
	local f = texture.orientation
	if f == "BOTH" then
		f = "HORIZONTAL"
	end
	self.drop:SetOrientation(f)

	self:SetStatusBarTexture(self:GetMediaPath('statusbar', texture.file) or texture.file)

	if background.solid == true then
		self.drop:SetStatusBarTexture(self:GetMediaPath('statusbar', "Flat") or texture.file)
	else
		self.drop:SetStatusBarTexture(self:GetMediaPath('statusbar', texture.file) or texture.file)
	end
    self:SetRotatesTexture(texture.rotateTexture)
    self.drop:SetRotatesTexture(texture.rotateTexture)

	self.opac:SetAlpha(texture.opacity)
end

function widget:SetBorder()
	local border = self.sets.visibility.border

	local file = self:GetBorderPath(border.file)
	
	for i, b in pairs(self.border) do
		if border.enable then
			b:Show()
		else
			b:Hide()
		end
		b:ClearAllPoints()
		b:SetTexture(file)
	end

	local hpad, vpad = border.hpadding+43, border.vpadding + 16
	
	local y, z = -5, 5
	local t, b = "Top", "Bottom"

	
	local handle = self.borderHandler
	handle:ClearAllPoints()
	handle:SetPoint("TopLeft", self, -hpad, vpad)
	handle:SetPoint("BottomRight", self, hpad, -vpad)

	local v, h = 30, 50
	local v, h = border.vthickness, border.hthickness
	
	self.border.TL:SetPoint("TopLeft", handle)
	self.border.TL:SetPoint("BottomRight", handle, "TopLeft", h, -v)

	self.border.BL:SetPoint("BottomLeft", handle)
	self.border.BL:SetPoint("TopRight", handle, "BottomLeft", h, v)

	self.border.L:SetPoint("TopLeft", self.border.TL, "BottomLeft")
	self.border.L:SetPoint("BottomRight", self.border.BL, "TopRight")

	self.border.TR:SetPoint("TopRight", handle, "TopRight")
	self.border.TR:SetPoint("BottomLeft", handle, "TopRight", -h, -v)

	self.border.BR:SetPoint("BottomRight", handle)
	self.border.BR:SetPoint("TopLeft", handle, "BottomRight", -h, v)

	self.border.R:SetPoint("TopLeft", self.border.TR, "BottomLeft")
	self.border.R:SetPoint("BottomRight", self.border.BR, "TopRight")	

	
	self.border.T:SetPoint("TopLeft", self.border.TL, "TopRight")
	self.border.T:SetPoint("BottomRight", self.border.TR, "BottomLeft")

	self.border.B:SetPoint("TopLeft", self.border.BL, "TopRight")
	self.border.B:SetPoint("BottomRight", self.border.BR, "BottomLeft")
	
	self.border.C:SetPoint("TopLeft", self.border.TL, "BottomRight")
	self.border.C:SetPoint("BottomRight", self.border.BR, "TopLeft")
--]]

	local i, j = 256, 64

	local Top_Start = 0
	local Bottom_Stop = 1
	
	local CH_Start, CH_Stop = 37/256, 1 - (37/256)
	
	local CV_Start, CV_Stop = 25/64, 1- (25/64)

	local Left_Start = 0
	local Right_Stop = 1

--[[	if border.flipVertical then
		b, t = "Top", "Bottom"
		x,y,z = -x , -y, -z
	end
--]]
	--left, right, top, bottom
	self.border.TL:SetTexCoord(Left_Start, CH_Start,   Top_Start, CV_Start)
	self.border.L :SetTexCoord(Left_Start, CH_Start,   CV_Start,  CV_Stop)
	self.border.BL:SetTexCoord(Left_Start, CH_Start,   CV_Stop,   Bottom_Stop)

	self.border.T:SetTexCoord( CH_Start,   CH_Stop,    Top_Start, CV_Start)
	self.border.C:SetTexCoord( CH_Start,   CH_Stop,    CV_Start,  CV_Stop)
	self.border.B:SetTexCoord( CH_Start,   CH_Stop,    CV_Stop,   Bottom_Stop)

	self.border.TR:SetTexCoord( CH_Stop,    Right_Stop, Top_Start, CV_Start)
	self.border.R:SetTexCoord( CH_Stop,    Right_Stop, CV_Start,  CV_Stop)
	self.border.BR:SetTexCoord( CH_Stop,    Right_Stop, CV_Stop,   Bottom_Stop)	
end


function widget:Update()
	if self.noUpdate then
		return
	end
	if self.OnUpdate then
		self:OnUpdate()
	end
end

function widget:SetColor(r, g, b, a)
	if ( r ~= self.r or g ~= self.g or b ~= self.b) or (a ~= self.a) then
		self:SetStatusBarColor(r, g, b)
		self.r, self.g, self.b, self.a = r, g, b, a
	end
end

function widget:GetMediaPath(kind, fileName)
	if Addon.lib then
		return Addon.lib:Fetch(kind, fileName)
	end
end

function widget:GetBorderPath(mediaName)
	return (Addon.lib and Addon.lib:Fetch("castborder", mediaName))
end



function widget:OnUpdate(elapsed)
	if self.noUpdate then
		return
	end

	self:UpdateCurrentCast(elapsed)
end



widget.Options = {

	{
		name = "Basic",
		kind = "Panel",
		key = "basic",
		panel = "Basic",
		options = {

			{
				name = 'Scale',
				kind = 'Slider',
				key = 'scale',
				min = 25,
				max = 200,
				panel = 'size',
			},
			{
				name = 'Width',
				kind = 'Slider',
				key = 'width',
				min = 10,
				max = 200,
				panel = 'size',
			},
			{
				name = 'Height',
				kind = 'Slider',
				key = 'height',
				min = 10,
				max = 200,
				panel = 'size',
			},
			{
				name = 'X Offset',
				kind = 'Slider',
				key = 'x',
				panel = 'position',
				min = -400,
				max = 400,
			},
			{
				name = 'Y Offset',
				kind = 'Slider',
				key = 'y',
				panel = 'position',
				min = -400,
				max = 400,
			},
			{
				name = 'Anchor',
				kind = 'Menu',
				key = 'anchor',
				panel = 'position',
				table = {
					'TopLeft',
					'Top',
					'TopRight',
					'Right',
					'BottomRight',
					'Bottom',
					'BottomLeft',
					'Left',
					'Center',
				},
			},
		{
			name = "Frame Level",
			kind = "Slider",
			key = "frameLevel",
			panel = 'position',
			min = 1,
			max = 100,
		},
		{
			name = "Frame Strata",
			kind = "Slider",
			key = "frameStrata",
			panel = 'position',
			min = 1,
			max = 7,
		},
			{
				name = 'Enable',
				kind = 'CheckButton',
				key = 'enable',
				panel = "advanced",
			},
			{
				name = 'Tooltip',
				kind = 'CheckButton',
				key = 'tooltip',
				panel = "advanced",
			},
		}
	},	
	{
 
		name = "Text",
		kind = "Panel",
		key = "text",
		panel = "Text",
		options = {
			{
				name = 'Enable',
				kind = 'CheckButton',
				key = 'enable',
				panel = "text",
			},
			{
				name = 'Font',
				kind = 'Media',
				key = 'file',
				mediaType = 'Font',
				panel = 'text',
			},
			{
				name = 'Size',
				kind = 'Slider',
				key = 'size',
				min = 1,
				max = 25,
				panel = 'size',
			},
			{
				name = 'Width',
				kind = 'Slider',
				key = 'width',
				min = 1,
				max = 150,
				panel = 'size',
			},
			{
				name = 'Height',
				kind = 'Slider',
				key = 'height',
				min = 1,
				max = 150,
				panel = 'size',
			},
			{
				name = 'Color',
				kind = 'ColorPicker',
				key = 'color',
				panel = 'text',
			},
			{
				name = 'X Offset',
				kind = 'Slider',
				key = 'x',
				panel = 'position',
				min = -400,
				max = 400,
			},
			{
				name = 'Y Offset',
				kind = 'Slider',
				key = 'y',
				panel = 'position',
				min = -400,
				max = 400,
			},
			{
				name = 'Justify Horizontal',
				kind = 'Slider',
				key = 'justifyH',
				panel = 'text',
				min = 1,
				max = 3,
			},
			{
				name = 'Justify Vertical',
				kind = 'Slider',
				key = 'justifyV',
				panel = 'text',
				min = 1,
				max = 3,
			},
			{
				name = 'Anchor',
				kind = 'Menu',
				key = 'anchor',
				panel = 'position',
				table = {
					'TopLeft',
					'Top',
					'TopRight',
					'Right',
					'BottomRight',
					'Bottom',
					'BottomLeft',
					'Left',
					'Center',
				},
			},
		}
	},
	{
 
		name = "Time",
		kind = "Panel",
		key = "time",
		panel = "Time",
		options = {
			{
				name = 'Enable',
				kind = 'CheckButton',
				key = 'enable',
				panel = "text",
			},
			{
				name = 'Font',
				kind = 'Media',
				key = 'file',
				mediaType = 'Font',
				panel = 'text',
			},
			{
				name = 'Size',
				kind = 'Slider',
				key = 'size',
				min = 1,
				max = 25,
				panel = 'text',
			},
			{
				name = 'Color',
				kind = 'ColorPicker',
				key = 'color',
				panel = 'text',
			},
			{
				name = 'X Offset',
				kind = 'Slider',
				key = 'x',
				panel = 'position',
				min = -400,
				max = 400,
			},
			{
				name = 'Y Offset',
				kind = 'Slider',
				key = 'y',
				panel = 'position',
				min = -400,
				max = 400,
			},
			{
				name = 'Anchor',
				kind = 'Menu',
				key = 'anchor',
				panel = 'position',
				table = {
					'TopLeft',
					'Top',
					'TopRight',
					'Right',
					'BottomRight',
					'Bottom',
					'BottomLeft',
					'Left',
					'Center',
				},
			},
		}
	},
	{
		name = "visibility",
		kind = "Panel",
		key = "visibility",
		panel = "visibility",
		options = {
			{
				name = 'Rotate Texture',
				kind = 'CheckButton',
				key = 'rotateTexture',
				panel = "texture",
			},
			{
				name = 'Fill Style',
				kind = 'Menu',
				key = 'fillstyle',
				panel = 'texture',
				table = {
					'REVERSE',
					'STANDARD_NO_RANGE_FILL',
					'STANDARD',
					'CENTER',
				},
			},
			
			{
				name = 'Texture',
				kind = 'Media',
				key = 'file',
				mediaType = 'statusbar',
				panel = 'texture',
			},
			{
				name = 'Opacity',
				kind = 'Slider',
				key = 'opacity',
				min = 0,
				max = 100,
				panel = 'texture',
			},
			{
				name = 'Orientation',
				kind = 'Menu',
				key = 'orientation',
				panel = 'texture',
				table = {
					'HORIZONTAL',
					'VERTICAL',
					'BOTH',
				},
			},
			{
				name = 'Enable',
				kind = 'CheckButton',
				key = 'enable',
				panel = "background",
			},
			{
				name = 'Background Color',
				kind = 'ColorPicker',
				key = 'color',
				panel = 'background',
			},

					{
						name = 'texture',
						kind = 'Media',
						key = 'file',
						mediaType = 'Castborder',
						panel = "border",
					},
			
                    {
                        name = 'Enable',
                        kind = 'CheckButton',
                        key = 'enable',
						panel = "border",
                    },
                    {
                        name = 'Solid Texture',
                        kind = 'CheckButton',
                        key = 'solid',
						panel = "background",
                    },
			{
				name = 'Vertical Padding',
				kind = 'Slider',
				key = 'vpadding',
				min = -50,
				max = 50,
				panel = 'border',
			},			
			{
				name = 'Horizontal Padding',
				kind = 'Slider',
				key = 'hpadding',
				min = -50,
				max = 50,
				panel = 'border',
			},
			{
				name = 'Vertical Thickness',
				kind = 'Slider',
				key = 'vthickness',
				min = 1,
				max = 50,
				panel = 'border',
			},
			{
				name = 'Horizontal Thickness',
				kind = 'Slider',
				key = 'hthickness',
				min = 1,
				max = 50,
				panel = 'border',
			},
			{
				name = 'Y',
				kind = 'Slider',
				key = 'y',
				min = -150,
				max = 150,
				panel = 'spark',
			},			
			{
				name = 'Curve',
				kind = 'Slider',
				key = 'offset',
				min = -150,
				max = 150,
				panel = 'spark',
			},
		}
	},

	}

--]]